package org.jboss.cache.marshall;

import org.jgroups.Channel;
import org.jgroups.MembershipListener;
import org.jgroups.Message;
import org.jgroups.MessageListener;
import org.jgroups.blocks.RpcDispatcher;

/**
 * Extends {@link org.jgroups.blocks.RpcDispatcher} and adds the possibility that the marshaller may throw {@link org.jboss.cache.marshall.InactiveRegionException}s.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
public class InactiveRegionAwareRpcDispatcher extends RpcDispatcher
{
   org.jboss.cache.marshall.Marshaller requestMarshaller;

   /**
    * Only provide the flavour of the {@link RpcDispatcher} constructor that we care about.
    */
   public InactiveRegionAwareRpcDispatcher(Channel channel, MessageListener l, MembershipListener l2, Object server_obj)
   {
      super(channel, l, l2, server_obj);
   }

   @Override
   public void setRequestMarshaller(Marshaller m)
   {
      super.setRequestMarshaller(m);
      requestMarshaller = (org.jboss.cache.marshall.Marshaller) m;
   }


   /**
    * Message contains MethodCall. Execute it against *this* object and return result.
    * Use MethodCall.invoke() to do this. Return result.
    */
   @Override
   public Object handle(Message req)
   {
      org.jgroups.blocks.MethodCall method_call;

      if (server_obj == null)
      {
         if (log.isErrorEnabled()) log.error("no method handler is registered. Discarding request.");
         return null;
      }

      if (req == null || req.getLength() == 0)
      {
         if (log.isErrorEnabled()) log.error("message or message buffer is null");
         return null;
      }

      RegionalizedMethodCall rmc;

      try
      {
         // we will ALWAYS be using the marshaller to unmarshall requests.
         rmc = requestMarshaller.regionalizedMethodCallFromByteBuffer(req.getBuffer());
         method_call = rmc.call;
      }
      catch (Throwable e)
      {
         if (e instanceof InactiveRegionException)
         {
            if (log.isTraceEnabled()) log.trace("Exception from marshaller: " + e.getMessage());
            return null;
         }

         if (log.isErrorEnabled()) log.error("exception unmarshalling object", e);
         return e;
      }

      try
      {
         if (log.isTraceEnabled())
            log.trace("[sender=" + req.getSrc() + "], method_call: " + method_call);

         Object retVal = method_call.invoke(server_obj);
         return new RegionalizedReturnValue(retVal, rmc);
      }
      catch (Throwable x)
      {
         return x;
      }
   }

   @Override
   public String toString()
   {
      return getClass().getSimpleName() + "[Outgoing marshaller: " + req_marshaller + "; incoming marshaller: " + rsp_marshaller + "]";
   }
}
